#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <cuda_runtime.h>

#define NX 8192
#define NY 8192
#define NT 100

#define BS 16

float data[2][NY][NX];
__device__ float gdata[2][NY][NX];

/* in milliseconds (ms) */
double get_elapsed_time(struct timeval *begin, struct timeval *end)
{
    return (end->tv_sec - begin->tv_sec) * 1000
            + (end->tv_usec - begin->tv_usec) / 1000.0;
}

void init()
{
  int jx, jy;
  
  for(jy = 0; jy < NY; jy++) {
    for(jx = 0; jx < NX; jx++) {
      float v = 0.0;
      if (jx+jy < (NX+NY)/4) {
	v = 1.0;
      }
      data[0][jy][jx] = v;
      data[1][jy][jx] = v;
    }
  }
  return;
}

/* Calculate for one time step */
/* Input: gdata[from], Output: gdata[1-from] */
__global__ void calc(int from)
{
  int to = 1-from;
  int jx = blockIdx.x * blockDim.x + threadIdx.x;
  int jy = blockIdx.y * blockDim.y + threadIdx.y;
  
  if (jx < 1 || jx >= NX-1 || jy < 1 || jy >= NY-1) {
    return;
  }

  gdata[to][jy][jx] = (float)0.2 * (gdata[from][jy][jx]
			     + gdata[from][jy][jx-1]
			     + gdata[from][jy][jx+1]
			     + gdata[from][jy-1][jx]
			     + gdata[from][jy+1][jx]);
  return;
}

int  main(int argc, char *argv[])
{
  int jt;
  void *gdata_ptr;
  struct timeval t1, t2;

  init();

  {
    int devid;
    cudaGetDevice(&devid);
    fprintf(stderr, "using GPU %d\n", devid);
  }

  gettimeofday(&t1, NULL);

  /* copy data on CPU to gdata on GPU */
  cudaGetSymbolAddress(&gdata_ptr, "gdata");
  cudaMemcpy(gdata_ptr, data, sizeof(float)*2*NX*NY, 
	     cudaMemcpyHostToDevice);

  {
    int from = 0;
    for (jt = 0; jt < NT; jt++) {
      fprintf(stderr, "step %d\n", jt);
      /* call GPU kernel function */
      calc<<<dim3((NX+BS-1)/BS, (NY+BS-1)/BS),
	dim3(BS, BS)>>>(from);
      cudaDeviceSynchronize();

      from = 1-from;
    }
  }

  /* copy gdata on GPU to data on CPU */
  cudaMemcpy(data, gdata_ptr, sizeof(float)*2*NX*NY, 
	     cudaMemcpyDeviceToHost);

  gettimeofday(&t2, NULL);
  printf("Elapsed time: %.1f msec\n", get_elapsed_time(&t1, &t2));
  
  return 0;
}
